
# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Kernel ridge regression

Penalized non-parametric regression through L2-regularized kernel
regression(aka ridge/Tykhonov regularization). It's often important
to regularize kernel regression due to the large number of parameters.

"""


import numpy as np
from sklearn.preprocessing import StandardScaler

from nystrompca.base import KernelMachine, Regression
from nystrompca.utils import to_column_vector, get_inverse


class KernelRR(KernelMachine, Regression):

    """
    Kernel ridge regression

    Parameters
    ----------
    gamma : float
        Ridge parameter

    """
    def __init__(self, gamma: float = 1e-14, **kwargs):

        super().__init__(**kwargs)

        self.gamma = gamma

        self.scaler = StandardScaler()


    def _setup(self, X: np.ndarray,
                  y: np.ndarray) -> None:
        """
        Preparations to fit the method

        Including required intial transformations of the X and y
        variables. Used by this method and Nyström KRR.

        """
        self.n = X.shape[0]

        self.X = self.scaler.fit_transform(X)

        y = to_column_vector(y)

        self.y_bar = y - np.mean(y)

        self.alpha = np.mean(y) # type: ignore[assignment]


    def fit(self, X: np.ndarray,
                  y: np.ndarray) -> None:
        """
        Please see parent class 'nystrompca.base.regression.Regression'
        for documentation

        """
        self._setup(X, y)

        self.K = self.kernel.matrix(self.X)

        M = self.gamma + self.K
        self.beta = get_inverse(M) @ self.y_bar


    def predict(self, X_new: np.ndarray) -> np.ndarray:
        """
        Please see parent class 'nystrompca.base.regression.Regression'
        for documentation

        Raises
        ------
        ValueError
            If the 'fit' method has not been called yet

        """
        if self.beta is None:
            raise ValueError("Call 'fit' before this function.")

        X_new = self.scaler.transform(X_new)

        kappa = self.kernel.matrix(self.X, X_new)
        predictions = self.alpha + kappa.T @ self.beta

        return predictions




"""
Implementation of Nyström Kernel PCA and a confidence bound on its accuracy

Experiments to evaluate the methods and confidence bound

Also implements a few other methods used for comparison

"""

from .algorithms.kernel_RR      import KernelRR
from .algorithms.kernel_PCA     import KernelPCA
from .algorithms.nystrom_KPCA   import NystromKPCA
from .algorithms.nystrom_KPCR   import NystromKPCR
from .algorithms.nystrom_KRR    import NystromKRR
from .algorithms.kernel_PCR     import KernelPCR
from .algorithms.bound          import calculate_bound, calc_conf_bound
from .experiments import (methods_experiments, bound_experiments,
                          regression_experiments)




# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Console script entry point to run the different experiments

"""

import argparse

from nystrompca.experiments import (methods_experiments, bound_experiments,
                                    regression_experiments)
from nystrompca.utils import logger


def main():
    """
    Entry point for command-line utility.

    Parses the arguments supplied to the command-line script and runs
    the selected experiment. Captures any exception that occurs and
    just prints the error message.

    """
    description = "Run the different Nyström kernel PCA experiments."
    epilog="Display subcommand options with nystrompca <subcommand> -h"
    parser = argparse.ArgumentParser(description=description,
                                     epilog=epilog)
    subparsers = parser.add_subparsers(required=True,
                                       dest='subcommand',
                                       title="available subcommands",
                                       help="which experiment to run")

    methods_description = methods_experiments.parser.description
    subparser_methods = subparsers.add_parser('methods',
                                       description= methods_description,
                                       parents=[methods_experiments.parser],
                                       add_help=False)
    subparser_methods.set_defaults(func=methods_experiments.main)

    bound_description = bound_experiments.parser.description
    subparser_bound = subparsers.add_parser('bound',
                                       description=bound_description,
                                       parents=[bound_experiments.parser],
                                       add_help=False)
    subparser_bound.set_defaults(func=bound_experiments.main)

    reg_description = regression_experiments.parser.description
    subparser_reg = subparsers.add_parser('regression',
                                       description=reg_description,
                                       parents=[regression_experiments.parser],
                                       add_help=False)
    subparser_reg.set_defaults(func=regression_experiments.main)

    args = vars(parser.parse_args())

    args.pop('subcommand')

    func = args.pop('func')

    try:
        func(**args)
    except Exception as e: # pylint: disable=broad-except
        logger.error(str(e))





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Create logging utility to be used by the rest of the code.

Creates a logger object which outputs to stdout with minimial formatting
and adds two new log levels, **header** and **subheader**, which have
additional formatting to demarcate sections of output.

Usage
~~~~~

::

    from nystrompca.utils import logger

    logger.header("Header log message")
    logger.subheader("Subheader log message")
    logger.info("Standard log message")

"""

import sys
import logging


HEADER    = 10
SUBHEADER = 15
logging.addLevelName(HEADER, 'HEADER')
logging.addLevelName(SUBHEADER, 'SUBHEADER')


def header(self, message):
    """
    New log level with customized formatting

    """
    self._log(HEADER, message, ())


def subheader(self, message):
    """
    New log level with customized formatting

    """
    self._log(SUBHEADER, message, ())


logging.Logger.header    = header
logging.Logger.subheader = subheader


class CustomFormatter(logging.Formatter):
    """
    Custom formatting for different log levels

    """

    def format(self, record):

        if record.levelno == HEADER:

            self._style._fmt = '\n %(message)s\n ' + '=' * 79

        elif record.levelno == SUBHEADER:

            self._style._fmt = '\n\n %(message)s\n ' + '-' * 79

        elif record.levelno == logging.INFO:

            self._style._fmt = '\n %(message)s'

        elif record.levelno == logging.WARN:
            self._style._fmt = '\n WARNING: %(message)s'

        elif record.levelno == logging.ERROR:
            self._style._fmt = '\n ERROR: %(message)s'

        formatted_record = logging.Formatter.format(self, record)

        return formatted_record


logger = logging.Logger('nystrompca')
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setFormatter(CustomFormatter())
logger.addHandler(stream_handler)
logger.setLevel(logging.DEBUG)





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Kernel principal component regression with the Nyström method.

"""

import numpy as np

from nystrompca import NystromKPCA
from nystrompca.base import Regression
from nystrompca.utils import to_column_vector


class NystromKPCR(NystromKPCA, Regression):
    """
    Implements Nyström kernel PCR

    """

    @staticmethod
    def from_pca(pca: NystromKPCA):
        """
        Create a Nyström PCR instance using the configuration
        from a Nyström PCA instance.

        """
        nystrom_kpcr = NystromKPCR(n_components = pca.n_components,
                                   m_subset     = pca.m,
                                   kernel       = pca.kernel.name,
                                   sigma        = pca.kernel.sigma,
                                   degree       = pca.kernel.degree,
                                   coef0        = pca.kernel.coef0,
                                   normalize    = pca.kernel.normalize)

        return nystrom_kpcr


    def fit(self, X: np.ndarray,
                  y: np.ndarray) -> None:
        """
        Fit the regression model.

        Parameters
        ----------
        X : numpy.ndarray, 2d
            Data matrix used for predictions
        y : numpy.ndarray, 1d or 2d
            True targets

        """
        y = to_column_vector(y)

        y_bar = np.mean(y)

        self.fit_transform(X)

        self.j = np.where(self.explained_variance_ > 0)[0][-1]
        M_inv = np.diag(1 / (self.explained_variance_[:self.j+1] * self.n))

        U_d = self.components_.T[:self.j+1]
        self.beta = M_inv @ U_d @ self.K_nm_p.T @ (y - y_bar)

        self.alpha = y_bar # type: ignore[assignment]


    def predict(self, X_new: np.ndarray) -> np.ndarray:
        """
        Please see parent class 'nystrompca.base.regression.Regression'
        for documentation

        Raises
        ------
        ValueError
            If the 'fit' method has not been called yet

        """
        if self.beta is None:
            raise ValueError("Call 'fit' before this function.")

        X_transformed = self.transform(X_new)

        predictions = self.alpha + X_transformed[:,:self.j+1] @ self.beta

        return predictions





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Kernel principal components analysis with the Nyström method.

"""

from typing import Tuple

import numpy as np

from nystrompca import KernelPCA
from nystrompca.base import NystromMethod
from nystrompca.utils import (get_inverse, get_eigendecomposition,
                              flip_dimensions)


class NystromKPCA(KernelPCA, NystromMethod):

    """
    Implements Nyström kernel PCA

    """
    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.K_mm_p: np.ndarray = None
        self.K_mm:   np.ndarray = None
        self.K_nm_p: np.ndarray = None
        self.K_nm:   np.ndarray = None


    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Calculate the Nyström kernel PCA for the supplied data matrix.

        """
        self._setup(X)

        if self.subset is None:
            self.create_subset(self.n)

        K_mm_p, K_mm, K_nm_p, K_nm = self.get_kernel_matrices()

        K_inv_sqrt = get_inverse(K_mm_p, np.sqrt)

        nystrom_matrix = K_inv_sqrt @ K_nm_p.T @ K_nm_p @ K_inv_sqrt / self.n
        M, V = get_eigendecomposition(nystrom_matrix)

        components_ = K_inv_sqrt @ V[:,:self.n_components]
        scores_     = K_nm_p @ components_

        scores_, components_ = flip_dimensions(scores_, components_)

        self.all_variances       = M.copy()
        self.explained_variance_ = M[:self.n_components]
        self.components_         = components_
        self.scores_             = scores_
        self.K_mm_p              = K_mm_p
        self.K_mm                = K_mm
        self.K_nm_p              = K_nm_p
        self.K_nm                = K_nm

        return self.scores_


    def transform(self, X_new: np.ndarray) -> np.ndarray:
        """
        Transform data into the coordinate system defined by the
        approximate Nyström principal components

        Raises
        ------
        ValueError
            If the 'fit_transform' method has not been called yet

        """
        if self.components_ is None:
            raise ValueError("Call 'fit_transform' before this function.")

        X_new = self.scaler.transform(X_new)

        kappa = self.kernel.matrix(self.X[self.subset], X_new, demean=False)

        kappa_p = self.get_kappa_p(kappa)

        X_transformed = kappa_p.T @ self.components_

        return X_transformed


    def get_reconstruction_errors(self, approx: bool = False) -> np.ndarray:
        """
        Calculate the reconstruction errors of the dataset onto
        the Nyström principal components.

        This method requires contruction of the full kernel matrix,
        be mindful that this is compute intensive. There is an option
        to create an approximate error which takes O(nm^2).

        Parameters
        ----------
        approx : bool
            Whether to approximate the mean of the full kernel matrix.

        Returns
        -------
        errors_ : numpy.ndarray
            Reconstruction errors

        Raises
        ------
        ValueError
            If the 'fit_transform' method has not been called yet

        """
        n = self.n
        X = self.X

        if self.components_ is None:
            raise ValueError("Call 'fit_transform' before this function.")

        if not self.demean:
            tot_var = np.sum([self.kernel(X[i], X[i]) for i in range(n)]) / n

        elif not approx:
            if self.K_p is None:
                self.K_p = self.kernel.matrix(X, demean=True)
            tot_var = np.trace(self.K_p) / n

        else:
            K_trace = np.array([self.kernel(X[i], X[i]) for i in range(n)])
            row_means = self.K_nm.sum(1) / len(self.subset)
            full_mean = row_means.sum() / n
            tot_var = np.sum(K_trace - 2 * row_means + full_mean) / n

        self.errors_ = tot_var - np.cumsum(self.explained_variance_)

        return self.errors_


    def get_subset_errors(self) -> np.ndarray:
        """
        Calculate the reconstruction error of all n data points onto
        the eigenspace from the subset of m subsampled points

        Returns
        -------
        errors : numpy.ndarray, 1d
            Reconstruction errors for all PCA dimensions

        Raises
        ------
        ValueError
            If the 'fit_transform' method has not been called yet

        """
        if self.components_ is None:
            raise ValueError("Call 'fit_transform' before this function.")

        if self.K_p is None:
            self.K_p = self.kernel.matrix(self.X, self.X, demean=self.demean)

        L_m, U_m = get_eigendecomposition(self.K_mm_p)

        j = np.where(L_m > 0)[0][-1]

        projections = np.zeros(self.n_components)

        for i in range(self.n_components):

            k = min(i,j)
            L_k = np.diag(1 / L_m[:k+1])
            U_k = U_m[:,:k+1]
            mat = self.K_nm_p @ U_k @ L_k @ U_k.T @ self.K_nm_p.T
            projections[i] = np.trace(mat)

        errors = np.trace(self.K_p) / self.n - projections / self.n

        return errors


    def get_subset_variance(self, X_new:  np.ndarray,
                                  PC_idx: int       ) -> float:
        """
        Calculate the amount of variance from new data captured
        by a Subset PCA principal components.

        Parameters
        ----------
        X_new : numpy.ndarray, 2d
            Data variance to capture
        PC_idx : int
            Which PC to consider

        Returns
        -------
        subset_variance: float
            Variance captured by one principal component

        Raises
        ------
        ValueError
            If 'fit_transform' has not been called yet

        """
        if self.K_mm_p is None:
            raise ValueError("Call 'fit_transform' before using this method.")

        kappa = self.kernel.matrix(self.X[self.subset], X_new, demean=False)
        kappa_p = self.get_kappa_p(kappa)

        L, U = get_eigendecomposition(self.K_mm_p)
        M = U[:,PC_idx].T @ kappa_p @ kappa_p.T @ U[:,PC_idx]
        eigenvalue_norm = X_new.shape[0] * L[PC_idx]
        if np.abs(eigenvalue_norm) < 1e-12:
            subset_variance = 0
        else:
            subset_variance = M / eigenvalue_norm

        return subset_variance


    def get_kernel_matrices(self) -> Tuple[np.ndarray,
                                           np.ndarray,
                                           np.ndarray,
                                           np.ndarray]:
        """
        Calculate the kernel matrices K_mm and K_nm

        Parameters
        ----------
        X: numpy.ndarray, 2d
            Data matrix
        subset: numpy.ndarray
            Subset indices
        kernel: nystrom_KPCA.kernel.Kernel
            Kernel wrapper class

        Returns
        -------
        K_mm_p : numpy.ndarray, 2d
            Centred kernel matrix with size m x m
        K_nm_p : numpy.ndarray, 2d
            Centred kernel matrix with size n x m
        K_nm : numpy.ndarray, 2d
            Kernel matrix with size n x m

        """
        K_nm = self.kernel.matrix(self.X, self.X[self.subset], demean=False)

        K_mm = K_nm[self.subset]

        if self.demean:
            K_mm_p, K_nm_p = demean_matrices(K_mm, K_nm)

        else:
            K_mm_p = K_mm
            K_nm_p = K_nm

        return K_mm_p, K_mm, K_nm_p, K_nm


    def get_kappa_p(self, kappa):
        """
        Get kappa as used in the nystrom demeaning

        """
        if not self.demean:
            kappa_p = kappa

        else:
            n, m = self.K_nm.shape
            n_new = kappa.shape[1]

            n_mean = self.K_nm.sum(0) / n

            M1 = np.tile(n_mean[:,np.newaxis], (1, n_new))
            m0 = n_mean @ get_inverse(self.K_mm)
            M2 = np.tile(m0 @ kappa, (m,1))
            M3 = n_mean @ m0

            kappa_p = kappa - M1 - M2 + M3

        return kappa_p




#############
### UTILS ###
#############


def demean_matrices(K_mm, K_nm):
    """
    Demean the matrices K_mm and K_nm

    """
    n, m = K_nm.shape

    n_mean = K_nm.sum(0) / n

    M1 = np.tile(n_mean, (n,1))

    m0 = get_inverse(K_mm) @ n_mean[:,np.newaxis]

    M2 = np.tile(K_nm @ m0, (1, m))

    M3 = n_mean @ m0

    K_nm_p = K_nm - M1 - M2 + M3

    M1 = M1[:m]

    K_mm_p = K_mm - M1 - M1.T + M3

    return K_mm_p, K_nm_p




# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Kernel ridge regression with the Nyström method

"""

import numpy as np

from nystrompca import KernelRR
from nystrompca.base import NystromMethod
from nystrompca.utils import get_inverse


class NystromKRR(KernelRR, NystromMethod):

    """
    Implements Nyström kernel ridge regression.

    """
    def __init__(self, **kwargs):

        super().__init__(**kwargs)


    def fit(self, X: np.ndarray,
                  y: np.ndarray) -> None:
        """
        Please see parent class 'nystrompca.base.regression.Regression'
        for documentation

        """
        self._setup(X, y)

        if self.subset is None:
            self.create_subset(self.n)

        K_nm = self.kernel.matrix(self.X, self.X[self.subset])
        K_mm = K_nm[self.subset]

        M = K_nm.T @ K_nm + self.gamma * K_mm
        self.beta = get_inverse(M) @ K_nm.T @ self.y_bar


    def predict(self, X_new: np.ndarray) -> np.ndarray:
        """
        Please see parent class 'nystrompca.base.regression.Regression'
        for documentation

        Raises
        ------
        ValueError
            If the 'fit' method has not been called yet

        """
        if self.beta is None:
            raise ValueError("Call 'fit' before this function.")

        X_transformed = self.scaler.transform(X_new)

        kappa = self.kernel.matrix(self.X[self.subset], X_transformed)
        predictions = self.alpha + kappa.T @ self.beta

        return predictions





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Calculation of the confidence bound for the
empirical reconstruction error

"""

import numpy as np
from scipy.stats import norm

from nystrompca.base import Kernel
from nystrompca.utils import get_eigendecomposition


def calculate_bound(X:      np.ndarray,
                    n:      int,
                    kernel: str,
                    alpha:  float     = 0.9,
                    **kwargs               ) -> np.ndarray:
    """
    Convenience function to calculate the confidence bound
    from a subset dataset and a kernel name.

    Parameters
    ----------
    X : numpy.ndarray, 2d
        A subset of data points
    kernel : str
        Kernel name
    n : int
        Total size of dataset
    alpha: float
        Confidence level
    **kwargs
        Kernel parameters (please see `nystrompca.base.Kernel`)

    Returns
    -------
    bounds : numpy.ndarray, 1d
        Confidence bound for all PCA dimensions

    """
    kernel_obj = Kernel(kernel=kernel, **kwargs)

    K_mm = kernel_obj.matrix(X, demean=False)

    bounds = calc_conf_bound(K_mm, n, kernel_obj.get_bound(), alpha=alpha)

    return bounds


def calc_conf_bound(K_mm : np.ndarray,
                    n:     int,
                    B:     float,
                    alpha: float     = 0.9) -> np.ndarray:
    """
    Calculation of confidence bound.

    Calculate a high-probability confidence bound on the empirical
    reconstruction error for kernel PCA with the Nyström method for
    all PCA dimensions.

    Parameters
    ----------
    K_mm : numpy.ndarray, 2d
        Subset kernel matrix
    n : int
        Total size of dataset
    B: float
        maximum value of kernel function
    alpha: float
        confidence level

    Returns
    -------
    bounds : numpy.ndarray, 1d
        Confidence bound for all PCA dimensions

    """
    m = K_mm.shape[0]

    if m == n: # Zero error for the Nyström method
        return np.zeros(m)

    L_m, _ = get_eigendecomposition(K_mm / m)

    eig_diff = L_m[:-1] - L_m[1:]

    delta = np.log(2 / (1 - alpha))

    term1   = B * np.sqrt(2 * delta) / np.sqrt(n-m)
    term2_1 = np.sqrt(2 * np.log(2))
    term2_2 = 2 * np.sqrt(2 * np.pi) * norm.cdf(-np.sqrt(2*np.log(2)))
    term2   = B**2 / np.sqrt(m) * (term2_1 + term2_2)
    D = (n - m) / n * (term1 + term2)

    D_k = np.ones(len(eig_diff))
    non_zero = eig_diff > 1e-14
    D_k[non_zero] = D**2 / eig_diff[non_zero]**2
    D_k[D_k > 1] = 1

    max_D = np.array([np.max(D_k[:i+1]) for i in range(m-1)])
    bounds = np.cumsum(L_m[:-1] * D_k) + D * max_D

    # Add NaN for the last PCA dimension
    bounds = np.r_[bounds, np.nan]

    return bounds





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Base class for different kernel methods. This class stores the kernel
function which is a required argument and the kernel matrix when it
is calculated.

"""

from abc import ABC

import numpy as np

from nystrompca.base import Kernel


class KernelMachine(ABC):

    """
    Abstract base class for kernel machines

    Parameters
    ----------
    kernel : str, default='rbf'
        Kernel function

    sigma : float, default=10
        Scale parameter for rbf, laplace and cauchy kernels

    degree: int, default=2
        Degree for the polynomial and rbf kernels

    coef0 : float, default=1
        Intercept for polynomial kernel

    normalize : bool, default=True
        Whether to normalize the polynomial kernel

    Attributes
    ----------
    kernel : nystrompca.algorithms.base.Kernel
        Kernel function instance

    K : numpy.ndarray, 2d
        Kernel matrix

    """
    def __init__(self, kernel:       str           = 'rbf',
                       sigma:        float         = 3,
                       degree:       int           = 2,
                       coef0:        float         = 1,
                       normalize:    bool          = True,
                       **kwargs                           ):

        super().__init__(**kwargs) # type: ignore[call-arg]

        self.kernel = Kernel(kernel, sigma, degree, coef0, normalize)

        self.K: np.ndarray = None





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Base class for the Nyström method.

"""

import numpy as np


class NystromMethod:

    """
    Base class for algorithms using the Nyström method

    Parameters
    ----------
    m_subset : int, default 10
        Size of Nyström subset
    seed : int, default None
        Seed for the random number generator

    """
    def __init__(self, m_subset: int = 10,
                       seed:     int = None,
                       **kwargs            ):

        super().__init__(**kwargs) # type: ignore[call-arg]

        self.m = m_subset

        self.subset: np.ndarray = None

        self.seed = seed


    def create_subset(self, n: int) -> None:
        """
        Randomly select a subset of *m* integers without replacement from
        all integers up to *n*. Sorts the selected subset in increasing order.

        Parameters
        ----------
        n : int
            Total dataset size

        Raises
        ------
        ValueError
            If *m* > *n*

        """
        rng = np.random.default_rng(self.seed)

        subset = rng.choice(range(n), self.m, replace=False)

        self.subset = np.sort(subset) # Not necessary but easier to debug




# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Experiments to illustrate Nyström kernel PCR and comparing it to
Nyström kernel ridge regression.

"""

# Built-in modules
import os
import sys
import argparse
from typing import Union, Tuple
from itertools import product

# Third-party packages
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

# This package
from nystrompca import NystromKPCR, NystromKRR
from nystrompca.experiments import (base_parser, get_airfoil_data,
                                    plot_R_squared, plot_targets)
from nystrompca.utils import logger


def main(kernel: str   = 'rbf',
         n:      int   = 1000,
         m:      int   = 100,
         d:      int   = 10,
         gamma:  float = 1,
         seed:   int   = None,
         noplot: bool  = False) -> int:
    """
    Run experiments for Nyström kernel PCR.

    Parameters
    ----------
    kernel : str, default 'rbf'
        Kernel function to use
    n : int, default 1000
        Total size of dataset
    m : int, default 100
        Size of subsampled dataset
    d : int, default 10
        Maximum subspace dimension
    gamma : float, default 1
        Ridge regularization parameter
    seed : int, default None
        Random seed for Nyström subset and train/test split
    noplot : bool, default False
        If True then don't display plots

    Returns
    -------
    int
        Return code, 0 if success

    """
    if noplot:
        os.environ['NYSTROMPCA_NOPLOT'] = "NOPLOT"

    logger.header("Regression experiments")

    logger.info(f"n: {n}")

    X, y = get_airfoil_data(n)

    X_train, X_test, y_train, y_test = train_test_split(X, y,
                                                        random_state=seed)

    run_multiple_regression_experiment(X_train, X_test, y_train, y_test,
                                       kernel, seed)

    run_single_regression_experiment(X_train, X_test, y_train, y_test,
                                     kernel, m, d, gamma, seed)

    return 0


def run_multiple_regression_experiment(X_train: np.ndarray,
                                       X_test:  np.ndarray,
                                       y_train: np.ndarray,
                                       y_test:  np.ndarray,
                                       kernel:  str       ,
                                       seed:    int       ) -> None:
    """
    Calculate and plot the R-squared across different parameters values.

    Calculate the regression for different subset sizes m and
    PCA dimensions d and plot the R-squared in a heat map.

    """
    logger.subheader("Experiment I: Multiple regressions R-squared")

    m_values = np.arange(10, 20)
    d_values = np.arange(11, 20)

    R_squared = pd.DataFrame(np.zeros((len(d_values), len(m_values))),
                             index=d_values, columns=m_values)

    for (m, d) in product(m_values, d_values):

        nystrom_kpcr = NystromKPCR(kernel=kernel, n_components=d,
                                   m_subset=m, sigma=1, seed=seed)

        R_squared.loc[d, m], _ = run_regression(nystrom_kpcr,
                                                X_train, X_test,
                                                y_train, y_test)

    R_squared.sort_index(ascending=False, inplace=True)

    plot_R_squared(R_squared, "Nyström KPCR", "PCA dimension")

    g_values = 1 / np.power(10, np.arange(5, 15))
    R_squared = pd.DataFrame(np.zeros((len(g_values), len(m_values))),
                             index=np.log10(g_values), columns=m_values)

    for (m, g) in product(m_values, g_values):

        nystrom_krr = NystromKRR(kernel=kernel, m_subset=m,
                                 gamma=g, sigma=1, seed=seed)

        R_squared.loc[np.log10(g), m] , _ = run_regression(nystrom_krr,
                                                           X_train, X_test,
                                                           y_train, y_test)

    plot_R_squared(R_squared, "Nyström KRR", "log of ridge parameter")


def run_single_regression_experiment(X_train: np.ndarray,
                                     X_test:  np.ndarray,
                                     y_train: np.ndarray,
                                     y_test:  np.ndarray,
                                     kernel:  str       ,
                                     m:       int       ,
                                     d:       int       ,
                                     gamma:   float     ,
                                     seed:    int       ) -> None:
    """
    Plot the predictions for one regression.

    Run one regression and plot the predicted target values versus
    the actual ones.

    """
    logger.subheader("Experiment II: Single regression predictions")

    kpcr = NystromKPCR(kernel=kernel, n_components=d, m_subset=m,
                       sigma=1, seed=seed)

    krr = NystromKRR(kernel=kernel, m_subset=m, gamma=gamma,
                     sigma=1, seed=seed)

    R_squared, predictions = run_regression(kpcr, X_train, X_test,
                                                  y_train, y_test)

    logger.info(f"R squared KPCR: {R_squared:.3f}")

    plot_targets(y_test, predictions, "Nyström KPCR", True)

    R_squared, predictions = run_regression(krr, X_train, X_test,
                                                 y_train, y_test)


    logger.info(f"R squared KRR: {R_squared:.3f}")

    plot_targets(y_test, predictions, "Nyström KRR", False)


def run_regression(model:   Union[NystromKPCR, NystromKRR],
                   X_train: np.ndarray,
                   X_test:  np.ndarray,
                   y_train: np.ndarray,
                   y_test:  np.ndarray) -> Tuple[float, np.ndarray]:
    """
    Run one regression and calculate the R squared and predictions

    """
    model.fit(X_train, y_train)

    predictions = model.predict(X_test)

    R_squared = model.score(X_test, y_test)

    return R_squared, predictions


description = "Run experiments for Nyström kernel PCR."
parser = argparse.ArgumentParser(description=description,
                                 parents=[base_parser])

parser.add_argument('-g', '--gamma', default=1, type=float,
                    help="ridge parameter")


if __name__ == '__main__':

    args = parser.parse_args()

    sys.exit(main(**vars(args)))





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Plotting functions for the output of the experiments.

All text is in the *Times New Roman* font

"""

import os
from pathlib import Path

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt


sns.set_context("paper")
sns.set(style="darkgrid")
sns.set_style({'font.family':'serif', 'font.serif':'Times New Roman'})



def plot_results(results: pd.DataFrame) -> None:
    """
    Plot the results of the PCA experiments.

    Creates four line plots with the values for the bound,
    the difference between the Nyström and standard reconstruction
    error, and the difference between the subset PCA reconstruction
    error and the true one.

    Parameters
    ----------
    results : pandas.DataFrame
        Experimental results. The data is divided into different
        plots by the **dataset** column

    """
    if 'NYSTROMPCA_NOPLOT' in os.environ.keys():
        return

    plot_vars = ['Nyström diff.','Subset diff.','Conf. bound']

    results = pd.melt(results[['d','dataset'] + plot_vars], ['d','dataset'])

    grid = sns.FacetGrid(results, col='dataset', col_wrap=2, hue='variable',
                         hue_kws={'color':  [(0.208, 0.388, 0.517),
                                             (0.213, 0.481, 0.673),
                                             (0.498, 0.498, 0.498)],
                                  'marker': ['o','s','.'],
                                  'ls':     ['-', '--', ':']})

    grid.map(plt.plot, 'd', 'value', alpha=0.8)

    grid.add_legend(title="\n", loc='upper right')
    grid.set_axis_labels("PCA dimension", "Errors")
    plt.show()


def plot_R_squared(R_squared: pd.DataFrame,
                   title:     str       = "",
                   ylabel:    str       = "") -> None:
    """
    Plot the R-squared from the regression experiments.

    Parameters
    ----------
    R_squared : pandas.DataFrame
        R-squared from multiple regressions
    title : str
        Plot title
    ylabel : str
        Label for y-axis

    """
    if 'NYSTROMPCA_NOPLOT' in os.environ.keys():
        return

    sns.set(font_scale=1.5)
    sns.set_style({'font.family':'serif', 'font.serif':'Times New Roman'})
    plt.gcf().subplots_adjust(bottom=0.15)
    plt.gcf().subplots_adjust(left=0.2)

    sns.heatmap(R_squared, cmap="YlGnBu")
    plt.title(title)
    plt.xlabel("Subset size")
    plt.ylabel(ylabel)
    plt.show()


def plot_targets(y:           np.ndarray,
                 predictions: np.ndarray,
                 title:       str       = "",
                 legend:      bool      = True) -> None:
    """
    Plot the predictions versus the actual targets for one regression
    in a heat map.

    Parameters
    ----------
    y : numpy.ndarray, 1d
        Actual targets
    predictions : numpy.ndarray, 1d
        Predicted targets
    title : str
        Plot title
    legend : bool, default=True
        Whether to draw a legend

    """
    if 'NYSTROMPCA_NOPLOT' in os.environ.keys():
        return

    y = y.squeeze()
    predictions = predictions.squeeze()

    sns.set(font_scale=1.5)
    sns.set_style({'font.family':'serif', 'font.serif':'Times New Roman'})
    plt.gcf().subplots_adjust(bottom=0.15)
    plt.gcf().subplots_adjust(left=0.15)

    # Same axis limits every time we create the plot.
    plt.ylim(102.5, 142.5)
    plt.xlim(102.5, 142.5)

    xy = np.array([y.min(), y.max()])
    plt.plot(xy, xy, color='gray', alpha=1, linewidth=3, linestyle='--')
    sns.scatterplot(x=y, y=predictions, s=100)
    plt.xlabel("Actual target")
    plt.ylabel("Predicted target")
    plt.title(title)

    if legend:
        plt.legend(['Oracle'])

    plt.show()





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Kernel principal components regression.

"""

import numpy as np

from nystrompca import KernelPCA
from nystrompca.base import Regression
from nystrompca.utils import to_column_vector


class KernelPCR(KernelPCA, Regression):

    """
    Implements kernel PCR

    """
    def __init__(self, **kwargs):

        super().__init__(**kwargs)


    def fit(self, X: np.ndarray,
                  y: np.ndarray) -> None:
        """
        Fit the regression model.

        Parameters
        ----------
        X : numpy.ndarray, 2d
            Data matrix used for predictions
        y : numpy.ndarray, 1d or 2d
            True targets

        """
        y = to_column_vector(y)

        y_bar = np.mean(y)

        self.fit_transform(X)

        j = np.where(self.explained_variance_ > 0)[0][-1]
        Lambda_inv = np.diag(1 / (self.n * self.explained_variance_[:j+1]))

        self.beta = np.sqrt(Lambda_inv) @ self.Q[:,:j+1].T @ (y - y_bar)

        self.beta = np.r_[self.beta, np.zeros((self.n_components-(j+1),1))]

        self.alpha = y_bar # type: ignore[assignment]


    def predict(self, X_new: np.ndarray) -> np.ndarray:
        """
        Create predictions from the fitted model.

        Parameters
        ----------
        X_new: numpy.ndarray, 1d or 2d
            New test points to create predictions for. Either 1d or
            a 2d row vector for one data point, or a matrix with multiple
            data points in the rows.

        Returns
        -------
        predictions : numpy.ndarray, 2d
            A column vector of predictions

        Raises
        ------
        ValueError
            If the 'fit' method has not been called yet

        """
        if self.beta is None:
            raise ValueError("Call 'fit' before this function.")

        X_transformed = self.transform(X_new)

        predictions = self.alpha + X_transformed @ self.beta

        return predictions


    def score(self, X: np.ndarray,
                    y: np.ndarray) -> float:
        """
        Calculate the regression R-squared.

        Parameters
        ----------
        X : numpy.ndarray, 2d
            Data matrix used for predictions
        y : numpy.ndarray, 1d or 2d
            True targets

        Returns
        -------
        R_squared : float
            R-squared

        """
        y = to_column_vector(y)

        predictions = self.predict(X)

        RSS = np.sum((y - predictions)**2)

        TSS = np.sum((y - y.mean())**2)

        R_squared = 1 - RSS / TSS

        return R_squared





"""
Base classes for the different methods in nystrompca.algorithms

"""

from .kernel         import Kernel
from .transformation import Transformation
from .regression     import Regression
from .kernel_machine import KernelMachine
from .nystrom_method import NystromMethod




# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Experiments comparing Nyström kernel PCA with other
unsupervised learning methods

* Linear PCA
* Kernel PCA
* Sparse PCA
* Locally Linear Embedding
* Independent Component Analysis

As a measure of the accuracy we use the average reconstruction errors
of a hold-out data set by default.

"""

# Built-in modules
import os
import sys
import time
import argparse
import warnings
from typing import Tuple

# Third-party packages
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.decomposition import PCA, SparsePCA, FastICA
from sklearn.manifold import LocallyLinearEmbedding

# This package
from nystrompca import KernelPCA, NystromKPCA
from nystrompca.experiments import base_parser, data
from nystrompca.utils import logger


# Always display DataFrame numbers in decimal format, using 4 decimals
pd.options.display.float_format = '{:,.4f}'.format


def main(kernel:   str  = 'rbf',
         n:        int  = 1000,
         m:        int  = 100,
         d:        int  = 10,
         ica:      bool = False,
         insample: bool = False,
         seed:     int  = None,
         noplot:   bool = False) -> int:
    """
    Run the experiments.

    Calculates the reconstruction error

    Parameters
    ----------
    kernel : str
        Kernel function to use
    n : int
        Total size of dataset
    m : int
        Size of subsampled dataset
    d : int
        Maximum subspace dimension
    ica : bool, default False
        If True then run ICA, which can be time-consuming
    insample : bool, default False
        If True then train and test the methods on the same data
    seed : int, default None
        Random seed for Nyström subset and train/test split
    noplot : bool, default False
        If True then don't display plots

    Returns
    -------
    int
        Return code, 0 if success

    """
    if noplot:
        os.environ['NYSTROMPCA_NOPLOT'] = "NOPLOT"

    logger.header("Methods comparison experiments")

    logger.info(f"Kernel: {kernel}")
    logger.info(f"n: {n}")
    logger.info(f"m: {m}")

    # Turn warnings into errors
    warnings.simplefilter('error')

    all_results = pd.DataFrame()

    t_nys_tot  = 0
    t_kpca_tot = 0

    for dataset in ('magic', 'yeast', 'cardiotocography', 'segmentation',
                    'drug', 'digits', 'dailykos', 'nips'):

        logger.subheader(f"Dataset: {dataset}")

        X = getattr(data, "get_" + dataset + "_data")(n)

        results, t_nys, t_kpca = run_one_dataset(kernel, X, m, d,
                                                 ica, insample, seed)
        t_nys_tot  += t_nys
        t_kpca_tot += t_kpca

        logger.info("Results:")
        logger.info(results)

        results['dataset'] = dataset

        all_results = all_results.append(results)

    logger.info(f"Kernel PCA elapsed time:  {t_kpca_tot / 8:.3f}")
    logger.info(f"Nyström PCA elapsed time: {t_nys_tot  / 8:.3f}")

    return 0


def run_one_dataset(kernel:   str,
                    X:        np.ndarray,
                    m:        int,
                    d:        int,
                    ica:      bool,
                    insample: bool,
                    seed:     int      ) -> pd.DataFrame:
    """
    Run an experiment for one kernel and dataset for all methods.

    Parameters
    ----------
    kernel : str
        Kernel function name
    X : numpy.ndarray, 2d
        Data matrix, with observations in the rows
    m : int
        The size of the Nyström subset
    d : int
        Subspace dimension
    ica : bool, default False
        If True then run ICA
    insample : bool, default False
        If True then don't use a test set
    seed : int, default None
        Random seed

    Returns
    -------
    results : pandas.DataFrame
        The results of one experiment

    """
    X_train, X_test = train_test_split(X, test_size=0.25, random_state=seed)

    if insample:
        X_train, X_test = X, X

    scaler  = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test  = scaler.transform(X_test)

    results = pd.DataFrame()

    (results['Subset PCA'],
     results['Nystrom PCA'],
     t_nys,
     sigma)                  = run_nystrom_kpca(X_train, X_test,
                                                kernel, m, d, seed)

    (results['Kernel PCA'],
     t_kpca)                 = run_kernel_pca(X_train, X_test,
                                              kernel, d, sigma)

    results['Linear PCA']    = run_linear_pca(X_train, X_test, d)

    results['Sparse PCA']    = run_sparse_pca(X_train, X_test, d)

    results['LLE']           = run_lle(X_train, X_test, d)

    if ica:
        results['ICA']       = run_ica(X_train, X_test, d)

    results['d'] = range(1,d+1)

    return results, t_nys, t_kpca


def run_nystrom_kpca(X_train: np.ndarray,
                     X_test:  np.ndarray,
                     kernel:  str,
                     m:       int,
                     d:       int,
                     seed:    int       ) -> Tuple[np.ndarray,
                                                   np.ndarray,
                                                   float,
                                                   float]:
    """
    Run Subset PCA and Nyström Kernel PCA experiment.

    Using the same subset for both methods for comparability.

    Parameters
    ----------
    X_train : numpy.ndarray, 2d
        Training data
    X_test : numpy.ndarray, 2d
        Validation data set, with the same number of data dimensions
        as the training set
    kernel: str
        kernel function name
    m : int
        Size of Nyström subset
    d : int
        Maximum number of components
    seed : int
        Random seed

    Returns
    -------
    fractions_subset : numpy.ndarray, 1d
        The fraction of explained variance for each dimension for Subset PCA
    fractions : numpy.ndarray, 1d
        The fraction of explained variance for each dimension for Nyström PCA
    t_nys : float
        The time taken to fit the method
    sigma : float
        The kernel bandwidth parameter, to use the same for standard KPCA

    """
    variances = np.zeros(d)

    nystrom_kpca = NystromKPCA(kernel=kernel, sigma=100, n_components=d,
                               m_subset=m, seed=seed)

    # Calculate bandwidth parameter based on a common heuristic
    nystrom_kpca.create_subset(X_train.shape[0])
    nystrom_kpca.kernel.calc_sigma(X_train[nystrom_kpca.subset])
    sigma = nystrom_kpca.kernel.sigma

    t0 = time.process_time()
    nystrom_kpca.fit_transform(X_train)
    t_nys = time.process_time() - t0

    variances_output = nystrom_kpca.transform(X_test).var(0)

    variances[:len(variances_output)] = variances_output # type: ignore[arg-type]

    # Subset PCA errors
    variances_subset = np.zeros(d)
    for i in range(d):
        variances_subset[i] = nystrom_kpca.get_subset_variance(X_test, i)

    # Calculate the fractions of the variance captured
    K = nystrom_kpca.kernel.matrix(X_test)

    total_variance = np.trace(K) / X_test.shape[0]

    fractions = np.cumsum(variances) / total_variance

    fractions_subset = np.cumsum(variances_subset) / total_variance

    return fractions_subset, fractions, t_nys, sigma


def run_kernel_pca(X_train: np.ndarray,
                   X_test:  np.ndarray,
                   kernel:  str,
                   d:       int,
                   sigma:   float     ) -> Tuple[np.ndarray, float]:
    """
    Run Kernel PCA experiment.

    Parameters
    ----------
    X_train : numpy.ndarray, 2d
        Training data
    X_test : numpy.ndarray, 2d
        Validation data set, with the same number of data dimensions
        as the training set
    kernel : str
        kernel function name
    d : int
        Maximum number of components
    sigma : float
        Kernel bandwidth parameter

    Returns
    -------
    fractions : numpy.ndarray, 1d
        The fraction of explained variance for each dimension

    """
    kernel_pca = KernelPCA(kernel=kernel, sigma=sigma, n_components=d)

    t0 = time.process_time()
    kernel_pca.fit_transform(X_train)
    t_kpca = time.process_time() - t0

    variances = kernel_pca.transform(X_test).var(0)

    K = kernel_pca.kernel.matrix(X_test, demean=True)

    total_variance = np.trace(K) / X_test.shape[0]

    fractions = np.cumsum(variances) / total_variance

    return fractions, t_kpca


def run_linear_pca(X_train: np.ndarray,
                   X_test:  np.ndarray,
                   d:       int       ) -> np.ndarray:
    """
    Run Linear PCA experiment.

    Parameters
    ----------
    X_train : numpy.ndarray, 2d
        Training data
    X_test : numpy.ndarray, 2d
        Validation data set, with the same number of data dimensions
        as the training set
    d : int
        Maximum number of components

    Returns
    -------
    fractions : numpy.ndarray, 1d
        The fraction of explained variance for each dimension

    """
    variances = np.zeros(d)

    d = min(d, X_train.shape[1])

    linear_pca = PCA(n_components=d)

    linear_pca.fit_transform(X_train)

    variances[:d] = linear_pca.transform(X_test).var(0)

    total_variance = np.sum(X_test.var(0)) # Sum of variances across dimensions

    fractions = np.cumsum(variances) / total_variance

    return fractions


def run_sparse_pca(X_train: np.ndarray,
                   X_test:  np.ndarray,
                   d:       int       ) -> np.ndarray:
    """
    Run Sparse PCA experiment.

    Sparse PCA is NP-hard in the number of components so we return
    NaNs if this is larger than 25.

    Parameters
    ----------
    X_train : numpy.ndarray, 2d
        Training data
    X_test : numpy.ndarray, 2d
        Validation data set, with the same number of data dimensions
        as the training set
    d : int
        Maximum number of components

    Returns
    -------
    fractions : numpy.ndarray, 1d
        The fraction of explained variance for each dimension

    """
    variances = np.zeros(d)
    variances.fill(np.nan)

    if X_train.shape[1] > 25:
        return variances

    d1 = min(d, X_train.shape[1])

    for i in range(d1):
        sparse_pca = SparsePCA(n_components=i+1)
        sparse_pca.fit_transform(X_train)
        variances[i] = sparse_pca.transform(X_test).var(0).sum()

    total_variance = np.sum(X_test.var(0))

    fractions = variances / total_variance

    fractions[np.isnan(fractions)] = max(fractions)

    return fractions


def run_lle(X_train: np.ndarray,
            X_test:  np.ndarray,
            d:       int       ) -> np.ndarray:
    """
    Run Locally-Linear Embedding experiment.

    The maximum possible number of components is the smallest of the
    number of data dimensions and one minus the number of data points.

    Parameters
    ----------
    X_train : numpy.ndarray, 2d
        Training data
    X_test : numpy.ndarray, 2d
        Validation data set, with the same number of data dimensions
        as the training set
    d : int
        Maximum number of components

    Returns
    -------
    fractions : numpy.ndarray, 1d
        The fraction of explained variance for each dimension

    """
    variances = np.zeros(d)
    variances.fill(np.nan)

    d = min(d, X_train.shape[1])

    for i in range(d):
        lle = LocallyLinearEmbedding(n_components=i+1)
        lle.fit(X_train)
        variances[i] = lle.transform(X_test).var(0).sum()

    # Find total variance by calculating the variance captured
    # by the maximum number of components
    max_dim = min(X_train.shape[0]-1, X_train.shape[1])
    lle = LocallyLinearEmbedding(n_components=max_dim)
    lle.fit(X_train)
    total_variance = lle.transform(X_test).var(0).sum()

    variances /= total_variance

    return variances


def run_ica(X_train: np.ndarray,
            X_test:  np.ndarray,
            d:       int       ) -> np.ndarray:
    """
    Run Indendent Components Analysis (ICA).

    For a too large number of components the algorithm will not converge,
    in which case we decrease the number of dimensions.

    Parameters
    ----------
    X_train : numpy.ndarray, 2d
        Training data
    X_test : numpy.ndarray, 2d
        Validation data set, with the same number of data dimensions
        as the training set
    d : int
        Maximum number of components

    Returns
    -------
    fractions : numpy.ndarray, 1d
        The fraction of explained variance for each dimension

    """
    variances = np.zeros(d)

    # ICA can not find more components than the number of data
    # dimensions or data points minus one
    max_dim = min(X_train.shape[0] - 1, X_train.shape[1])
    max_d   = min(max_dim, d)
    for i in range(max_d):
        ica = FastICA(n_components=i+1, tol=10e-3, max_iter=int(5000))
        try:
            ica.fit(X_train)
            variances[i] = ica.transform(X_test).var(0).sum()
        except: # pylint: disable=bare-except
            variances[i] = variances[i-1]

    # Find the total variance by calculating the variance captured with the
    # max number of components. It's hard to know what is the largest number
    # for the algorithm to converge, so we try different ones
    while 1:
        ica = FastICA(n_components=max_dim, tol=10e-3, max_iter=int(1000))
        try:
            ica.fit(X_train)
        except: # pylint: disable=bare-except
            max_dim -= 1
        else:
            break

    total_variance = ica.transform(X_test).var(0).sum()

    variances /= total_variance

    return variances


description = "Comparison of unsupervised methods."
parser = argparse.ArgumentParser(description=description,
                                 parents=[base_parser])

parser.add_argument('--ica', action='store_true',
                    help="include ICA, which can be time-consuming")

parser.add_argument('--insample', action='store_true',
                    help="train and test methods on the same data")


if __name__ == '__main__':

    args = parser.parse_args()

    sys.exit(main(**vars(args)))





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Abstract base class for dimensionality reduction methods with only
independent variables that calculates new data representations
in a different coordinate system.

"""

from abc import ABC, abstractmethod

import numpy as np


class Transformation(ABC):
    """
    Abstract base class for dimensionality reduction methods that
    transform the data.

    n_components : int, default None
        Number of latent dimensions to use, such as the number of
        principal components. Default is unspecified and depends on
        the particular method.

    Attributes
    ----------
    scores_ : numpy.ndarray, 2d
        Transformed data points, with each new data point in a row
        of the array. These are for example the principal scores.
        They should be created in the 'fit_transform' method.

    """
    def __init__(self, n_components: int = None, **kwargs):

        super().__init__(**kwargs) # type: ignore[call-arg]

        self.n_components:   int = n_components

        self.X:       np.ndarray = None

        self.n:              int = None

        self.scores_: np.ndarray = None


    @abstractmethod
    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Fit the data transformation

        Parameters
        ----------
        X: numpy.ndarray, 2d
            Input data matrix with each data point in a row

        Returns
        -------
        scores_ : numpy.ndarray
            Transformed data points with new data points along the rows,
            such as the principal scores.

        """
        ...


    @abstractmethod
    def transform(self, X_new: np.ndarray) -> np.ndarray:
        """
        Transform data using the fitted model

        Parameters
        ----------
        X_new : numpy.ndarray, 2d
            Data to transform, with one data point in each row.

        Returns
        --------
        X_transformed : numpy.ndarray
            Transformed data with same size as input data

        """
        ...


    def tot_variance(self, X: np.ndarray = None):
        """
        Calculate the sum of the variances across all dimensions
        in the data. Using n degrees of freedom.

        Parameters
        ----------
        X : numpy.ndarray, 2d
            Dataset

        Raises
        ------
        ValueError
            If no dataset exists

        """
        if X is None:
            X = self.X

        if X is None:
            raise ValueError("No dataset supplied")

        total_variance = np.sum(X.var(0))

        return total_variance





"""
Experiments to illustrate Nyström kernel PCA, the confidence bound
and Nyström kernel PCR.

"""

from .plotting import plot_results, plot_R_squared, plot_targets
from .data import (get_magic_data, get_yeast_data, get_cardiotocography_data,
                   get_segmentation_data, get_airfoil_data)
from .base_parser import base_parser
from . import (methods_experiments, bound_experiments, regression_experiments,
               plotting, data)




# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module contains function to read different datasets from the
UCI machine learning repository. The raw dataset are included in
the repository in the `data/` folder in the root directory, apart
from the `digits` dataset which can be downloaded with scikit-learn.

"""

from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.datasets import load_digits


DATA_FOLDER = Path(__file__).parent.joinpath('../../data/')


def get_magic_data(n: int) -> np.ndarray:
    """
    Read the magic telescope data

    Parameters
    ----------
    n : int
        Maximum number of data points

    Returns
    -------
    data : numpy.ndarray
        Data matrix

    """
    data = []
    with open(DATA_FOLDER.joinpath("magic_gamma_telescope.dat")) as f:
        for line in f.readlines():
            split_line = line.split(',')[:-1]
            data.append(split_line)

    data_arr = np.asarray(data, dtype=np.float64)

    data_arr = data_arr[:n]

    return data_arr


def get_yeast_data(n: int) -> np.ndarray:
    """
    Read the yeast dataset with protein location sites for fungi

    Parameters
    ----------
    n : int
        Maximum number of data points

    Returns
    -------
    data : numpy.ndarray
        Data matrix

    """
    df = pd.read_csv(DATA_FOLDER.joinpath("yeast.dat"), header=None,
                     delimiter=r'\s+')

    df.drop(0, axis=1, inplace=True)

    df = pd.get_dummies(df)

    return df.values[:n]


def get_cardiotocography_data(n: int) -> np.ndarray:
    """
    Read the cardiotocography dataset with heart measurement data

    Parameters
    ----------
    n : int
        Maximum number of data points

    Returns
    -------
    data : numpy.ndarray
        Data matrix

    """
    data = []
    with open(DATA_FOLDER.joinpath("cardiotocography.dat")) as f:
        for line in f.readlines():
            line = line.split('\n')[0]
            split_line = line.split('\t')[3:] # keep numeric data
            data.append(split_line)

    data_arr = np.asarray(data, dtype=np.float64)

    data_arr = data_arr[:n]

    return data_arr


def get_segmentation_data(n: int) -> np.ndarray:
    """
    Read the segmentation dataset with various data on images

    Parameters
    ----------
    n : int
        Maximum number of data points

    Returns
    -------
    data : numpy.ndarray
        Data matrix

    """
    data = []
    with open(DATA_FOLDER.joinpath("segmentation.dat")) as f:
        for line in f.readlines():
            split_line = line.split(' ')
            data.append(split_line)

    data_arr = np.asarray(data, dtype=np.float64)

    data_arr = data_arr[:n]

    return data_arr


def get_drug_data(n: int) -> np.ndarray:
    """
    Read the drug consumption dataset

    Parameters
    ----------
    n : int
        Maximum number of data points

    Returns
    -------
    data : numpy.ndarray
        Data matrix

    """
    data = pd.read_csv(DATA_FOLDER.joinpath("drug.dat"),
                       header=None, index_col=0)

    data = data.replace('CL0', 0)
    data = data.replace('CL1', 1)
    data = data.replace('CL2', 2)
    data = data.replace('CL3', 3)
    data = data.replace('CL4', 4)
    data = data.replace('CL5', 5)
    data = data.replace('CL6', 6).astype(float)

    return data.values[:n]


def get_digits_data(n: int) -> np.ndarray:
    """
    Read the digits dataset from UCI through the scikit-learn helper
    function. The dataset contains the flattened grayscale pixel values
    from 8x8 images of handwritten digits.

    Parameters
    ----------
    n : int
        Maximum number of data points

    Returns
    -------
    X : numpy.ndarray, 2d
        Independent variables

    """
    digits = load_digits()

    return digits['data'][:n]


def get_dailykos_data(n: int) -> np.ndarray:
    """
    Read the dailykos dataset with word frequencies of blog posts
    into bag-of-words vectors

    Parameters
    ----------
    n : int
        Maximum number of bag-of-words vectors

    Returns
    -------
    X : numpy.ndarray, 2d
        Independent variables

    """
    df = read_bag_of_words('dailykos.dat')

    X = df.values[:n]

    return X


def get_nips_data(n: int) -> np.ndarray:
    """
    Read the dailykos dataset with word frequencies in NIPS papers
    into bag-of-words vectors

    Parameters
    ----------
    n : int
        Maximum number of bag-of-words vectors

    Returns
    -------
    X : numpy.ndarray, 2d
        Independent variables

    """
    df = read_bag_of_words('nips.dat')

    X = df.values[:n]

    return X


def read_bag_of_words(dataset: str) -> pd.DataFrame:
    """
    Read word frequencies and convert them into bag-of-words vectors

    Parameters
    ----------
    dataset: str
        name of file to read from

    Returns
    -------
    df : pandas.DataFrame
        Bag-of-words vectors along the rows

    """

    df = pd.read_csv(DATA_FOLDER.joinpath(dataset),
                     names=['docID','wordID','count'], delimiter=' ')

    df = df.pivot(index='docID', columns='wordID', values='count')

    df.replace(to_replace=np.nan, value=0, inplace=True)

    return df


def get_airfoil_data(n: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Read the airfoil dataset with wind tunnel measurements from NASA.
    This dataset is used in the regression experiments

    Parameters
    ----------
    n : int
        Maximum number of data points

    Returns
    -------
    X : numpy.ndarray, 2d
        Independent variables
    y : numpy.ndarray, 1d
        Dependent variable

    """
    data = []
    with open(DATA_FOLDER.joinpath("airfoil.dat")) as f:
        for line in f.readlines():
            split_line = line.split('\t')
            data.append(split_line)

    data_arr: np.ndarray = np.asarray(data, dtype=np.float64)

    data_arr = data_arr[:n]

    X: np.ndarray = data_arr[:,:-1]
    y: np.ndarray = data_arr[:,-1]

    return X, y





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Kernel principal component analysis.

"""

import numpy as np
from sklearn.preprocessing import StandardScaler

from nystrompca.base import KernelMachine, Transformation
from nystrompca.utils import (get_eigendecomposition, get_kappa,
                              IdentityScaler, demean_matrix, flip_dimensions)


class KernelPCA(KernelMachine, Transformation):

    """
    Implements Kernel PCA.

    Parameters
    ----------
    scale : bool, default=True
        Whether to scale the input data to zero mean and unit variance

    Attributes
    ----------
    explained_variance_ : numpy.ndarray, 1d
        The variances of data along the principal components. Created
        in the 'fit_transform' method.

    components_ : numpy.ndarray, 2d
        The principal components, in terms of weightings on the m
        data points, with one principal component in each column
        of the array. Created in the 'fit_transform' method.

    errors_ : numpy.ndarray, 1d
        Reconstruction error of the full dataset on the principal
        components. Each element in the array is the reconstruction error
        with respect to all principal components up to that index.
        Calculated in the 'calculate_errors' method.

    """
    def __init__(self, scale:  bool = True,
                       demean: bool = True,
                       **kwargs           ):

        super().__init__(**kwargs)

        if scale:
            self.scaler = StandardScaler()
        else:
            self.scaler = IdentityScaler()

        self.demean = demean

        self.explained_variance_: np.ndarray = None
        self.all_variances:       np.ndarray = None
        self.components_:         np.ndarray = None
        self.errors_:             np.ndarray = None
        self.K_p:                 np.ndarray = None
        self.Q:                   np.ndarray = None


    def _setup(self, X: np.ndarray) -> None:

        self.n = X.shape[0]

        self.X = self.scaler.fit_transform(X)

        if self.n_components is None:
            self.n_components = self.n


    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """
        Calculate the Kernel PCA for the supplied data matrix.

        Calculates the principal components, scores and values
        (explained variance).

        """
        self._setup(X)

        K = self.kernel.matrix(self.X, demean=False)

        if self.demean:
            K_p = demean_matrix(K)
        else:
            K_p = K

        L, Q = get_eigendecomposition(K_p / self.n)

        explained_variance_ = L[:self.n_components]

        Lambda = np.diag(self.n * explained_variance_)

        Q = Q[:,:self.n_components]
        scores_ = Q @ np.sqrt(Lambda)

        scores_, Q = flip_dimensions(scores_, Q)

        # Save calculated variables
        self.K                   = K
        self.K_p                 = K_p
        self.scores_             = scores_
        self.explained_variance_ = explained_variance_
        self.Q                   = Q

        return scores_


    def transform(self, X_new: np.ndarray) -> np.ndarray:
        """
        Transform data into the coordinate system defined by the
        principal components including demeaning the data.

        Raises
        ------
        ValueError
            If the 'fit_transform' method has not been called yet

        """
        if self.scores_ is None:
            raise ValueError("Call 'fit_transform' before this function.")

        X_new = self.scaler.transform(X_new)

        n_new = X_new.shape[0]

        kappa_p = self.kernel.matrix(self.X, X_new, demean=False)

        kappa_p = get_kappa(kappa_p, self.K, self.demean)

        j = np.where(self.explained_variance_ > 0)[0][-1]
        Lambda_inv = np.diag(1 / (self.n * self.explained_variance_[:j+1]))

        X_transformed = kappa_p.T @ self.Q[:,:j+1] @ np.sqrt(Lambda_inv)

        # Add zeros for dimensions with zero explained variance
        X_transformed = np.c_[X_transformed,
                              np.zeros((n_new, self.n_components-(j+1)))]

        return X_transformed





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Experiments to evaluate Nyström kernel PCA and the confidence bound,
comparing it to standard kernel PCA and kernel PCA using the subset
of data points directly.

"""

# Built-in modules
import os
import sys
import argparse
from typing import Tuple
from multiprocessing import Pool, cpu_count

# Third-party packages
import numpy as np
import pandas as pd
from sklearn.preprocessing import scale

# This package
from nystrompca import NystromKPCA, calc_conf_bound
from nystrompca.base import Kernel
from nystrompca.utils import logger, get_eigendecomposition, get_tail_sums
from nystrompca.experiments import base_parser, data, plot_results


# Always display numbers in decimal format, using 4 decimals
pd.options.display.float_format = '{:,.4f}'.format


def main(kernel:  str   = 'rbf',
         n:       int   = 1000,
         m:       int   = 100,
         d:       int   = 10,
         alpha:   float = 0.9,
         samples: int   = 1,
         seed:    int   = None,
         noplot: bool   = False) -> int:
    """
    Run the experiments four different datasets.

    After each plot is shown the program halts, close the plot to continue.

    Parameters
    ----------
    kernel: str
        Kernel function to use
    n : int
        Total size of dataset
    m : int
        Size of subsampled dataset
    d : int
        Maximum PCA dimension
    alpha : float
        Confidence level
    samples : int, default 1
        Number of subset samples to average
    seed : int, default None
        Random seed for the Nyström subset
    noplot: bool, default False
        If True then don't display plots

    Returns
    -------
    int
        Return code, 0 if success

    """
    if noplot:
        os.environ['NYSTROMPCA_NOPLOT'] = "NOPLOT"

    logger.header("Confidence bound experiments")

    logger.info(f"Kernel: {kernel}")
    logger.info(f"n: {n}")
    logger.info(f"m: {m}")
    logger.info(f"samples: {samples}")
    logger.info(f"alpha: {alpha}")

    all_results = pd.DataFrame()

    for dataset in ('magic', 'yeast', 'cardiotocography', 'segmentation'):

        logger.subheader(f"Dataset: {dataset}")

        X = getattr(data, "get_" + dataset + "_data")(n)

        results = run_one_experiment(kernel, X, m, d, alpha, samples, seed)
        logger.info("Results:")
        logger.info(results[['d',
                             'Nyström diff.',
                             'Subset diff.',
                             'Conf. bound',
                             'Total variance']])

        results['dataset'] = dataset

        all_results = all_results.append(results)

    all_results.reset_index(drop=True, inplace=True)

    plot_results(all_results)

    return 0


def run_one_experiment(kernel:  str,
                       X:       np.ndarray,
                       m:       int,
                       d:       int,
                       alpha:   float,
                       samples: int,
                       seed:    int       ) -> pd.DataFrame:
    """
    Run an experiment for one kernel and dataset.

    Calculate the Nyström PCA bound, the actual difference between the
    Nyström and full reconstruction errors, and the difference between
    the full reconstruction error and the reconstruction error for the
    PCA subspace from the subset directly, for a number of different
    PCA subspace dimensions.

    The functions calculates the Nyström errors and bounds for multiple
    different subset samples and then takes the average. This is done
    in parallel using the maximum number of available CPU cores.

    Parameters
    ----------
    kernel : str
        Kernel function
    X : numpy.ndarray, 2d
        Data matrix, with observations in the rows
    m : int
        The size of the Nyström subset
    d : int
        Maximum PCA dimension
    alpha : float
        Confidence level
    samples : int
        Number of samples over which to calculate the Nyström quantities
    seed : int
        Random seed

    Returns
    -------
    results : pandas.DataFrame
        The results of one experiment

    """
    if samples == 1:
        results = get_nystrom_results(X, kernel, m, d, alpha, seed)

    else:
        with Pool(cpu_count()) as p:
            results = p.starmap(get_nystrom_results,
                                samples*[(X, kernel, m, d, alpha, seed)])

        results = pd.concat(results, axis=0)

        results = results.groupby('d', as_index=False).mean()

    true_errors, total_variance = get_true_errors(X, kernel, d)

    results['True errors']    = true_errors
    results['Nyström diff.']  = results['Nyström errors'] - true_errors
    results['Subset diff.']   = results['Subset errors']  - true_errors
    results['Total variance'] = total_variance

    # Some values may be zero within machine epsilon, set these to exactly zero
    results[np.abs(results) < 1e-14] = 0

    return results


def get_nystrom_results(X:      np.ndarray,
                        kernel: str,
                        m:      int,
                        d:      int,
                        alpha:  float,
                        seed:   int     ) -> pd.DataFrame:
    """
    Get experimental results for one Nyström PCA experiment, for one
    sampling of the subset.

    Calculates the reconstruction errors, explained variances,
    subset PCA reconstruction errors and the confidence bounds
    for multiple PCA dimensions.

    Parameters
    ----------
    X : numpy.ndarray, 2d
        Data matrix
    kernel : str
        Name of kernel function, supplied to the NystromKPCA class.
    m : int
        Subset size
    d : int
        Maximum PCA dimension
    alpha : float
        Confidence level
    seed : int
        Random seed

    Returns
    -------
    one_result : pandas.DataFrame
        Table with results, with columns 'd', 'Nyström errors', 'Nyström PCA',
        'Subset errors', 'Conf. bound', and with each row containing the
        results for one PCA dimension

    """
    n = X.shape[0]

    nystrom_kpca = NystromKPCA(kernel       = kernel,
                               n_components = d,
                               m_subset     = m,
                               demean       = False,
                               seed         = seed)

    nystrom_kpca.fit_transform(X)

    one_result = pd.DataFrame()

    one_result['d']              = np.arange(d) + 1
    one_result['Nyström errors'] = nystrom_kpca.get_reconstruction_errors()
    one_result['Nyström PCA']    = get_tail_sums(nystrom_kpca.all_variances, d)

    one_result['Subset errors']  = nystrom_kpca.get_subset_errors()

    one_result['Conf. bound']    = calc_conf_bound(nystrom_kpca.K_mm_p, n,
                                               nystrom_kpca.kernel.get_bound(),
                                               alpha)[:d]

    return one_result


def get_true_errors(X:          np.ndarray,
                    kernelname: str,
                    d:          int       ) -> Tuple[np.ndarray, float]:
    """
    Calculate the reconstruction error for standard kernel PCA

    Parameters
    ----------
    X : numpy.ndarray, 2d
        Data matrix
    kernel : str
        Name of kernel function, supplied to the NystromKPCA class.
    d : int
        Maximum PCA dimension

    Returns
    -------
    true_errors : numpy.ndarray
        Standard kernel PCA reconstruction errors for all PCA dimensions
    total_variance : float
        The total "variance" of the data (reconstruction error for
        an empty subspace)

    """
    n = X.shape[0]

    kernel = Kernel(kernelname)

    K = kernel.matrix(scale(X), demean=False)
    L, _ = get_eigendecomposition(K / X.shape[0])

    # Reconstruction errors for PCA on the entire dataset (for all d)
    true_errors = get_tail_sums(L, d)

    # Maximum reconstruction error
    total_variance = np.trace(K) / n

    return true_errors, total_variance # type: ignore[return-value]


description = "Calculate the confidence bound for different datasets."
parser = argparse.ArgumentParser(description=description,
                                 parents=[base_parser])

parser.add_argument('--alpha', default=0.9, type=float,
                    help="confidence level")

parser.add_argument('--samples', default=1, type=int,
                    help="number of subset samples to average")


if __name__ == '__main__':

    args = parser.parse_args()

    sys.exit(main(**vars(args)))





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Class that encapsulates a kernel function and its parameters.

Contains functionality to evaluate a kernel function and calculate
a kernel matrix, and keeps track of the maximum kernel value.

"""

from itertools import product

import numpy as np

from nystrompca.utils import demean_matrix


class Kernel:

    """
    Wrapper for kernel functions.

    Please see 'nystrompca.base.kernel_machine' for argument descriptions

    """
    def __init__(self, kernel:    str   = 'rbf',
                       sigma:     float = 3,
                       degree:    int   = 2,
                       coef0:     float = 1,
                       normalize: bool  = True ):

        self.sigma        = sigma
        self.degree       = degree
        self.coef0        = coef0
        self.normalize    = normalize

        self.name = kernel

        bound: float

        if kernel == 'rbf':
            kernel_fcn = lambda x, y: rbf(x, y, degree, sigma)
            bound = 1

        elif kernel == 'poly':
            kernel_fcn = lambda x, y: poly(x, y, degree, coef0, normalize)
            if normalize:
                bound = 1
            else:
                bound = np.inf

        elif kernel == 'linear':
            kernel_fcn = lambda x, y: poly(x, y, 1, 0, normalize=False)
            bound = np.inf

        elif kernel == 'laplace':
            kernel_fcn = lambda x, y: rbf(x, y, 2, sigma)
            bound = 1

        elif kernel == 'cauchy':
            kernel_fcn = lambda x, y: cauchy(x, y, sigma)
            bound = 1

        elif kernel == 'cosine':
            kernel_fcn = lambda x, y: poly(x, y, 1, 0, normalize=True)
            bound = 1

        self.kernel_fcn = kernel_fcn

        self.bound = bound


    def __call__(self, x, y):
        return self.kernel_fcn(x, y)


    def get_name(self) -> str:
        """
        Getter for the kernel name

        """
        return self.name


    def get_bound(self) -> float:
        """
        Getter for the kernel function bound

        """
        return self.bound


    def calc_sigma(self, X: np.ndarray):
        """
        Calculate sigma based on a common heuristic

        Calculate the bandwidth parameter for the RBF and Cauchy kernels
        as the average distance between all pairs of data points

        Parameters
        ----------
        X : numpy.ndarray, 2d
            Input data

        """
        assert X.ndim == 2, "Two-dimensional input data expected"

        n = X.shape[0]

        pairwise_diff = np.tile(X, (n,1)) - X.repeat(n, axis=0)
        sigma = np.sqrt((pairwise_diff**2).sum(1)).sum() / n**2

        # Reinitialize with new sigma
        self.__init__(self.name, sigma, self.degree, # type: ignore[misc]
                      self.coef0, self.normalize)


    def matrix(self, X1:     np.ndarray,
                     X2:     np.ndarray = None,
                     demean: bool       = True) -> np.ndarray:
        """
        Calculate kernel matrix.

        Calculate the kernel matrices between two datasets.

        """
        n = X1.shape[0]

        # If only one dataset is supplied we only need to calculate half
        # the kernel evaluations explicitly plus the diagonal
        if X2 is None:
            K = np.zeros((n,n))
            for i in range(n):
                K[i, i] = self(X1[i], X1[i]) / 2
                for j in range(i+1, n):
                    K[i,j] = self(X1[i], X1[j])
            K = K + K.T

        # Otherwise we need to calculate all elements explicitly
        else:
            m = X2.shape[0]
            K = np.zeros((n,m))
            for i, j in product(range(n), range(m)):
                K[i,j] = self(X1[i],X2[j])

        if demean:
            K = demean_matrix(K)

        return K


def rbf(x:      np.ndarray,
        y:      np.ndarray,
        degree: int,
        sigma:  float     ) -> float:
    """
    Radial Basis Functions kernel

    """
    return np.exp(-np.sqrt((x - y) @ (x - y))**degree / (sigma**2))


def poly(x:         np.ndarray,
         y:         np.ndarray,
         degree:    int,
         coef0:     float,
         normalize: bool      = True) -> float:
    """
    Polynomial kernel

    """
    value = (x @ y + coef0)**degree
    if normalize:
        value /= ((x @ x + coef0) * (y @ y + coef0))**(degree/2)

    return value


def cauchy(x:     np.ndarray,
           y:     np.ndarray,
           sigma: float     ) -> float:
    """
    Cauchy kernel

    """
    return 1 / (1 + (x - y) @ (x - y) / sigma**2)





# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Base parser object for arguments used by all experiments


Usage
~~~~~

::

    import argparse

    from nystrompca.experiments import base_parser

    parser = argparse.ArgumentParser(parents=[base_parser])

"""

import argparse

base_parser = argparse.ArgumentParser(add_help=False)

base_parser.add_argument('-n', default=1000, type=int,
                         help="total size of dataset")

base_parser.add_argument('-m', default=100, type=int,
                         help="size of subsampled dataset")

base_parser.add_argument('-d', default=10, type=int,
                         help="PCA dimension")

base_parser.add_argument('--kernel', default='rbf',
                         choices=['rbf', 'poly', 'linear', 'cosine',
                                  'laplace', 'cauchy'],
                         help="kernel function")

base_parser.add_argument('--seed', default=None, type=int,
                         help="Random seed")

base_parser.add_argument('--noplot', action='store_true',
                         help="don't produce plots")






from setuptools import setup, find_packages


with open("requirements.txt") as f:
    requirements_list = f.read().split('\n')


setup(name                 = "nystrompca",
      version              = "1.0.1",
      description          = "Kernel PCA with the Nyström method",
      author               = "Fredrik Hallgren",
      author_email         = "fredrik.hallgren@ucl.ac.uk",
      url                  = "https://github.com/fredhallgren/nystrompca",
      packages             = find_packages(),
      long_description     = """This package implements an efficient non-linear PCA by combining kernel PCA with the Nyström randomized subsampling method, as well as a confidence bound on the accuracy of the method. """,
      long_description_content_type="text/plain",
      classifiers         = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"],
      python_requires      = ">=3.6",
      install_requires     = requirements_list,
      include_package_data = True,
      data_files = [('data', ['data/magic_gamma_telescope.dat',
                              'data/yeast.dat',
                              'data/cardiotocography.dat',
                              'data/segmentation.dat',
                              'data/nips.dat',
                              'data/drug.dat',
                              'data/dailykos.dat',
                              'data/airfoil.dat'])],
      entry_points         = {
          'console_scripts': [
              'nystrompca=nystrompca.experiments.console_script:main']})






# Copyright 2021 Fredrik Hallgren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Various utility functions.

"""

from typing import Tuple, Callable

import numpy as np


def demean_matrix(M: np.ndarray) -> np.ndarray:
    """
    Subtract the mean in feature space from a kernel matrix.

    Parameters
    ----------
    M : numpy.ndarray, 2d
        Matrix to be mean-adjusted

    Returns
    -------
    M_p : numpy.ndarray
        Adjusted matrix

    """
    m, n = M.shape

    M1 = M.sum(0).repeat(m).reshape(n,m).T / m
    M2 = M.sum(1).repeat(n).reshape(m,n) / n
    M3 = np.sum(M) / (m*n)
    M_p = M - M1 - M2 + M3

    return M_p


def get_eigendecomposition(M:   np.ndarray,
                           eps: float     = 1e-12
                           ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculate the eigendecomposition of a symmetric positive semi-definite
    matrix. Sets negative eigenvalues or eigenvalues smaller than "eps"
    to zero.

    Parameters
    ----------
    M: numpy.ndarray, 2d
        Symmetric positive semi-definite matrix

    Returns
    -------
    L : numpy.ndarray, 1d
        Eigenvalues, ordered in decreasing order
    U : np.ndarray, 2d
        Eigenvectors

    """
    L, U = np.linalg.eigh(M)
    L = L[::-1]
    U = U[:,::-1]

    # Sometimes the matrix can have small negative eigenvalues due to numerical
    # inaccuracies, despite being positive semidefinite in theory. Set these
    # to zero. Also remove tiny eigenvalues for numerical regularization.
    L[L < eps] = 0

    return L, U


def get_inverse(M:    np.ndarray,
                func: Callable  = lambda x: x,
                eps:  float     = 1e-9) -> np.ndarray:
    """
    Calculate the regularized inverse, removing small or negative eigenvalues
    before taking the inverse.

    Parameters
    ----------
    M : numpy.ndarray, 2d square
        matrix to be inverted
    func : Callable, default identity
        Transformation to apply to the eigenvalues before
        taking the inverse

    Returns
    -------
    M_inv : numpy.ndarray, square 2d
        Inverse matrix

    """
    L, U = get_eigendecomposition(M, eps)

    j = np.where(L > 0)[0][-1]
    M_inv = U[:,:j+1] @ np.diag(1 / func(L[:j+1])) @ U[:,:j+1].T

    return M_inv


def get_tail_sums(L: np.ndarray,
                  d: int       = None) -> np.ndarray:
    """
    Calculate the first d sums of the tail of an array.

    Parameters
    ----------
    L : numpy.ndarray, 1d
        Numbers to be summed
    d : int, optional
        Number of tail sums to compute. If None then compute all
        tail sums.

    Returns
    -------
    L_sums : np.ndarray
        Tail sums

    """
    if d is None:
        d = len(L)

    L_sums = (L.sum() - L.cumsum())[:d]

    return L_sums


def get_kappa(kappa:  np.ndarray,
              K:      np.ndarray,
              demean: bool      = True) -> np.ndarray:
    """
    Demean the kappa vector of kernel evaluations of new data points
    with the the existing ones. Works either for one new data point, in
    which case "kappa" should be a column vector, or for multiple data
    points with "kappa" a matrix.

    Parameters
    ----------
    kappa : numpy.ndarray, 2d
        Kernel evaluations of new data points
    K : numpy.ndarray, 2d
        Uncentred kernel matrix, with the same data points used in "kappa"

    Returns
    -------
    kappa_p
        Centred kernel evaluations

    """
    if not demean:
        kappa_p = kappa

    else:
        m, n = K.shape
        n_new = kappa.shape[1]

        kappa_mean = np.tile(kappa.sum(0), (m, 1)) / m

        n_mean = np.tile(K.sum(1)[:, np.newaxis], n_new) / n

        nm_mean = np.tile(K.sum() / (n * m), (m, n_new))

        kappa_p = kappa - kappa_mean - n_mean + nm_mean

    return kappa_p


def to_column_vector(y: np.ndarray) -> np.ndarray:
    """
    Take a one-dimensional vector or a row vector and
    convert into a column vector.

    Parameters
    ----------
    y : numpy.ndarray, 1d or 2d
        Vector to convert

    Returns
    -------
    y : numpy.ndarray, 2d
        Column vector

    """
    if y.ndim == 1:
        y = y[:, np.newaxis]

    if y.ndim == 2 and y.shape[0] == 1:
        y = y.T

    return y


def flip_dimensions(scores:     np.ndarray,
                    components: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    The principal components and scores are only unique up to a sign,
    so we flip the dimensions so that the range of scores in each dimenstion
    lies mostly within positive values. Specifically, we flip dimensions
    so that the mid point of the range of values is positive.

    Parameters
    ----------
    scores : numpy.ndarray, 2d
        Use these to calculate which dimensions to flip and flip them
    components : numpy.ndarray, 2d
        Flip these too

    Returns
    -------
    scores_flipped : numpy.ndarray, 2d
        Flipped vectors
    components_flipped : numpy.ndarray, 2d
        Flipped vectors

    """
    flip = (scores.min(0) + scores.max(0)) / 2 < 0
    flip_matrix = np.diag(1 - 2 * flip)

    scores_flipped     = scores     @ flip_matrix
    components_flipped = components @ flip_matrix

    return scores_flipped, components_flipped


class IdentityScaler:
    """
    This class is used to scale the input data points
    when no adjustment is desired and the input data is
    used as is.

    This class is used in the unit tests.

    """

    @classmethod
    def fit_transform(cls, X: np.ndarray) -> np.ndarray:
        """
        Do nothing

        """
        return X

    @classmethod
    def transform(cls, X: np.ndarray) -> np.ndarray:
        """
        Do nothing

        """
        return X

